integrate_likelihood_extrinsic --amp-order 0 --n-eff 1000 --time-marginalization --mass1 1.2199  --save-P 0.0001 --save-deltalnL inf --n-max 1000000 --output-file testme.xml.gz --convergence-tests-on True --n-chunk 1000 --save-samples --approximant TaylorT4 --reference-freq 0.0 --cache-file 2015_bns_mdc.cache --l-max 2 --mass2 1.20066 --adapt-floor-level 0.1 --adapt-weight-exponent 0.312986728635 --event-time 966822123.739409208 --fmax 2000.0 --psd-file H1=H1_PSD_measured.xml.gz --psd-file L1=L1_PSD_measured.xml.gz --channel-name H1=FAKE-STRAIN --channel-name L1=FAKE-STRAIN 
